#!/usr/bin/env python3
"""A script to install or uninstall the CloudLinux Subsystem for Ubuntu."""

from __future__ import annotations

import argparse
import logging
import os
import re
import select
import shutil
import subprocess
import sys
import tempfile
from enum import Enum
from functools import total_ordering
from glob import glob
from pathlib import Path
from subprocess import PIPE, STDOUT, Popen
from typing import IO
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen, urlretrieve

script_file = __file__
script_version = "1.0.4"
script_name = "ubuntu2cloudlinux.py"

# ruff will always complain about urllib.request.urlretrieve when S310 is enabled
# ruff: noqa: S310

VERSION_FILE_URL = "https://repo.cloudlinux.com/cloudlinux-ubuntu/sources/cln/version.dat"
SCRIPT_URL = "https://repo.cloudlinux.com/cloudlinux-ubuntu/sources/cln/ubuntu2cloudlinux.py"

SOURCES_DIR = "/etc/apt/sources.list.d"
LOG_PATH = "/var/log/ubuntu2cloudlinux.log"
ROOT_CPANEL_DIR = "/usr/local/cpanel/whostmgr/docroot/"

# ubuntu2cloudlinux supports these Ubuntu versions only
SUPPORTED_UBUNTU_VERSIONS = ["20.04", "22.04"]


class Color(Enum):
    """Terminal colors represented by ANSI escape codes."""

    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    NOCOLOR = "\033[0m"


class PanelName(Enum):
    """Control panel names."""

    CPANEL = "cpanel"
    NOPANEL = "nopanel"


@total_ordering
class ScriptVersion:
    """
    A class to parse and compare version strings.

    Version strings are expected to be in the format "x.y.z" where x, y, z are integers.
    This isn't PEP 440 compliant, but it's good enough for our purposes.

    There are already libraries that can handle version comparison, but since we ship this script
    as a standalone file, we don't want to require dependencies from the Python environment that runs it.
    """

    def __init__(self, version_str: str) -> None:
        """Parse version string into components."""
        parts = version_str.split(".")
        # Convert all parts to integers, default to 0 for invalid parts
        self.components = [int(part) if part.isdigit() else 0 for part in parts]

    def __eq__(self, other: ScriptVersion) -> bool:
        """Compare if this version is equal to other."""
        return self.components == other.components

    def __lt__(self, other: ScriptVersion) -> bool:
        """Compare if this version is less than other."""
        for s, o in zip(self.components, other.components):
            if s != o:
                return s < o
        return len(self.components) < len(other.components)

    def __ge__(self, other: ScriptVersion) -> bool:
        """Compare if this version is greater than or equal to other."""
        return not self.__lt__(other)


COMMON_PACKAGES = [
    "alt-python27-cllib",
    "app-version-detector",
    "cloudlinux-linksafe",
    "kmodlve-dkms",
    "lve",
    "lvemanager",
    "lve-stats",
    "lve-utils",
    "lve-wrappers",
    "pam-lve",
]
RHN_PACKAGES = [
    "rhn-check",
    "rhn-client-tools",
    "rhn-setup-gnome",
    "rhn-setup",
    "rhnlib",
    "rhnsd",
]
CL_SOLO_META_PACKAGE = "cloudlinux-solo-meta"
CL_ADMIN_META_PACKAGE = "cloudlinux-admin-meta"
CL_CONTAINER_META_PACKAGE = "cloudlinux-container"

# mod_hostinglimits packages have different naming schemes for different panels.
PANEL_PACKAGES_MAP = {
    PanelName.CPANEL: ["ea-apache24-mod-hostinglimits"],
    PanelName.NOPANEL: ["mod-hostinglimits"],
}


def setup_parser() -> argparse.Namespace:
    """Set up the argument parser."""
    global script_name

    parser = argparse.ArgumentParser(
        description="Installation/uninstallation script for the CloudLinux Subsystem for Ubuntu",
        prog="ubuntu2cloudlinux.py",
        usage="%(prog)s [options]",
    )

    registration_group = parser.add_mutually_exclusive_group(required=True)
    registration_group.add_argument(
        "-k",
        "--key",
        dest="activationkey",
        help="activation key for registration",
        default="",
    )
    registration_group.add_argument(
        "-i",
        "--byip",
        dest="activationkey",
        help="register by IP",
        action="store_const",
        const="",
    )
    registration_group.add_argument(
        "--skip-registration",
        help="skip registration",
        action="store_true",
    )

    parser.add_argument(
        "--serverurl",
        help="server URL for registration (default is %(default)s)",
        default="https://xmlrpc.cln.cloudlinux.com/XMLRPC",
    )

    parser.add_argument(
        "--skip-version-check",
        dest="skipversioncheck",
        help="skip version check of %(prog)s",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--skip-full-update",
        dest="skipfullupdate",
        help="skip the full package update before conversion",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--force-new-configs",
        dest="forcenewconf",
        help="force package installation by overwriting conflicting configs",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--uninstall",
        dest="mode_uninstall",
        help="uninstall CloudLinux Subsystem for Ubuntu",
        action="store_true",
        default=False,
    )

    script_name = parser.prog
    return parser.parse_args()


def setup_logger() -> logging.Logger:
    """Set up logging into stdout and ubuntu2cloudlinux.log file."""
    console_formatter = logging.Formatter("%(message)s")
    file_formatter = logging.Formatter("%(asctime)s - [%(levelname)s] - %(message)s")

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(logging.INFO)

    file_handler = logging.FileHandler(filename=LOG_PATH, mode="a")
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)

    logger = logging.getLogger()
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)
    logger.setLevel(logging.DEBUG)

    return logger


# Placed in the global scope to be accessible from everywhere in the script.
logger = setup_logger()


def log_with_color(
    message: str | list[str],
    color: Color,
    level: int = logging.INFO,
) -> None:
    """
    Log a message decorated with a specific terminal color.

    @param message: Message to log.
    @param color: Sequence of ANSI escape codes to set the terminal color.
    Will be reset to default after the message.
    @param level: Logging level.
    """
    # We accept both strings and lists of strings as input,
    # but for simplicity, just convert strings to a list if we get one.
    if isinstance(message, str):
        message = [message]

    for msg_line in message:
        logger.log(level, "%s%s%s", color.value, msg_line, Color.NOCOLOR.value)


def get_os_info() -> dict[str, str]:
    """Collect info about OS from /etc/os-release file."""
    os_info = {}
    with Path("/etc/os-release").open("r") as f:
        for line in f:
            param, value = line.strip().split("=")
            os_info[param] = value.strip('"')
    return os_info


def check_ubuntu() -> None:
    """
    Check that the script is running on a supported Ubuntu version.

    Exit if the script is run on an unsupported version or OS.
    """
    os_info = get_os_info()
    os_name, os_version = os_info["NAME"], os_info["VERSION_ID"]

    if os_name != "Ubuntu" or os_version not in SUPPORTED_UBUNTU_VERSIONS:
        logger.error("Failed to run %s script on %s %s.", script_name, os_name, os_version)
        logger.error(
            "It should only be used on the following Ubuntu versions: %s",
            ", ".join(SUPPORTED_UBUNTU_VERSIONS),
        )
        logger.error("Other Ubuntu versions are currently not supported.")
        sys.exit(1)


def get_panel_name() -> PanelName:
    """Determine control panel name."""
    if Path(ROOT_CPANEL_DIR).is_dir():
        return PanelName.CPANEL
    return PanelName.NOPANEL


def create_repo_file(filename: str, repo_config: str) -> None:
    """
    Create a file with the repository configuration inside.

    Writes a *.list file in the sources directory with the given repository config inside it.
    @param filename: Name of the file to create.
    @param repo_config: Repository configuration (as a string) to write into the file.
    """
    file_path = Path(SOURCES_DIR) / filename
    logger.info("Creating apt source file: %s", file_path)
    with Path(file_path).open("w") as f:
        f.write(repo_config)


def log_subprocess_output(pipe: IO[bytes]) -> None:
    """
    Log subprocess output from the pipe (typically stdout).

    Handle subprocess output as follows:
    - log all lines into logfile,
    - log line into stdout if it contains "Setting up" to make it clear to the user at what step he is.
    """
    line_to_search = "Setting up"
    for byte_line in iter(pipe.readline, b""):
        line = byte_line.decode().strip(" ").strip("\n")
        if line_to_search in line:
            logger.info(line)
        else:
            logger.debug(line)


def run_command(command: list[str], *, exit_on_error: bool = True) -> int:
    """
    Run a command in a subprocess.

    @param command: Command to execute.
    @param exit_on_error: Exit if the executed command a returned non-zero exit code.
    """
    env = os.environ.copy()
    # DEBIAN_FRONTEND=noninteractive to disable debconf interactive interface
    # This will make apt-get select the default options
    env.update({"DEBIAN_FRONTEND": "noninteractive"})

    # The command executed comes from this script, so there is no need to validate it
    process = Popen(command, stdout=PIPE, stderr=STDOUT, env=env)  # noqa: S603
    if process.stdout is None:
        logger.warning("No stdout for the process %s from command %s", process, command)
        return 1
    with process.stdout:
        log_subprocess_output(process.stdout)

    exitcode = process.wait()
    if exit_on_error and exitcode != 0:
        sys.exit(1)
    return exitcode


def install_from_url(url: str) -> None:
    """
    Perform package installation from a repository or custom URLs.

    @param url: Link to download and install the package from.

    @raises URLError: If an error occurs while downloading the package.
    """
    # Packages downloaded should be provided from cloudlinux.com.
    # Verify the URL to prevent downloading and installing packages from other sources.
    parsed_url = urlparse(url)
    # Root domain must be cloudlinux.com, but the subdomain can be anything
    if not parsed_url.netloc.endswith("cloudlinux.com"):
        logger.error("Invalid URL provided: %s", url)
        sys.exit(1)

    with tempfile.NamedTemporaryFile(dir="/root") as tmp:
        # urlretrieve can raise URLError, but handling it is the caller's responsibility
        urlretrieve(url, tmp.name)
        run_command(["dpkg", "-i", tmp.name])


def install_cl_release(os_version: str) -> None:
    """Install the cloudlinux-release package."""

    # Use direct URL if set. Useful for internal testing.
    direct_release_url = os.getenv("CLOUDLINUX_RELEASE_URL")
    if direct_release_url:
        cloudlinux_release_url = direct_release_url
    else:
        # Public repo URLs look like this:
        # https://repo.cloudlinux.com/cloudlinux-ubuntu/cloudlinux/stable/cloudlinux-release-latest-20_04.deb
        cloudlinux_release_url = (
            "https://repo.cloudlinux.com/cloudlinux-ubuntu/cloudlinux/stable/"
            f"cloudlinux-release-latest-{os_version}.deb"
        )

    try:
        install_from_url(cloudlinux_release_url)
    except URLError as err:
        logger.info("Error encountered while downloading and installing from URL %s: %s", cloudlinux_release_url, err)
        sys.exit(1)


def install_ea4_release(os_version: str) -> None:
    """Install the cloudlinux-ea4-release package."""

    # Use direct URL if set. Useful for internal testing.
    direct_release_url = os.getenv("CLOUDLINUX_EA4_RELEASE_URL")
    if direct_release_url:
        release_url = direct_release_url
    else:
        # Public repo URLs look like this:
        # https://repo.cloudlinux.com/cloudlinux-ubuntu/cloudlinux/stable/cloudlinux-ea4-release-latest-20_04.deb
        release_url = (
            "https://repo.cloudlinux.com/cloudlinux-ubuntu/cloudlinux-ea4/stable/"
            f"cloudlinux-ea4-release-latest-{os_version}.deb"
        )

    try:
        install_from_url(release_url)
    except URLError as err:
        logger.info("Error encountered while downloading and installing from URL %s: %s", release_url, err)
        sys.exit(1)


def setup_build_repos() -> None:
    """
    Set up CLBS repos for testing.

    - get build IDs from the environment variable;
    - download the deployment tool script;
    - enable builds from BS.
    """
    build_ids_str = os.getenv("CLDEPLOY_TEST_BUILD_IDS", "")
    if build_ids_str == "":
        return

    auth_token = os.getenv("CLDEPLOY_TEST_AUTH_TOKEN")
    if auth_token is None:
        # Build IDs are set, but auth token is not - make this notable
        logger.error("CLDEPLOY_TEST_AUTH_TOKEN is not set, unable to set up builds from BS")
        raise RuntimeError

    bs_deployment_tool_url = "https://build.cloudlinux.com/api/v1/deployment-tool/"
    deploy_dst_file = "/var/run/deploy.py"

    try:
        urlretrieve(bs_deployment_tool_url, deploy_dst_file)
    except URLError as err:
        logger.info("Error encountered while downloading and copying deployment tool: %s", err)
        sys.exit(1)

    for id_ in build_ids_str.split(","):
        run_command(["python3", deploy_dst_file, "--with-linked", "-t", auth_token, id_])
        logger.info("Enabled build from BS: %s", id_)


def register_server() -> None:
    """Perform registration of the system."""
    if parser_args.skip_registration:
        logger.info("Registration skipped as the --skip-registration option was passed")
        return

    logger.info("Starting registration of the server...")
    common_opts = [
        f"--serverUrl={parser_args.serverurl}",
        "--edition=auto",
        "--migrate-silently",
    ]

    if parser_args.activationkey:
        run_command([
            "/usr/sbin/rhnreg_ks",
            f"--activationkey={parser_args.activationkey}",
            *common_opts,
        ])
    else:
        run_command([
            "/usr/sbin/clnreg_ks",
            *common_opts,
        ])
    logger.info("Registration completed successfully")


def get_packages_to_install() -> list[str]:
    """Get a list of packages to install depending on the environment."""
    panel_name = get_panel_name()

    packages = [
        *COMMON_PACKAGES,
        *PANEL_PACKAGES_MAP.get(panel_name, []),
    ]
    if Path("/proc/vz/vestat").exists() or os.getenv("CLDEPLOY_TEST_MOCK_OPENVZ_MARKER") == "true":
        logger.info("OpenVZ environment detected - will install CloudLinux Container specific packages")
        packages.append(CL_CONTAINER_META_PACKAGE)

    return packages


def perform_ubuntu_conversion() -> None:
    """
    Perform the steps during the Ubuntu conversion process.

    The following steps are performed:
    - update all installed packages (if not skipped intentionally)
    - install cloudlinux-release to enable CloudLinux repositories
    - install package required for registration and register the server
    - install all other required packages
    - reconfigure CloudLinux to match the edition of the key is registered with
    """
    logger.info("Updating info about all available packages...")
    run_command(["apt-get", "update"])

    apt_confnew = ["-o", "Dpkg::Options::=--force-confnew"]
    apt_confdef = ["-o", "Dpkg::Options::=--force-confdef"]
    apt_confold = ["-o", "Dpkg::Options::=--force-confold"]
    if not parser_args.skipfullupdate:
        logger.info("Updating all currently installed packages...")
        if parser_args.forcenewconf:
            run_command(["apt-get", "upgrade", "-y", *apt_confnew])
        else:
            # with confdef+confold dpkg overwrites configuration files that have not been modified,
            # but for modified configs the new version is installed with a .dpkg-dist suffix
            run_command(["apt-get", "upgrade", "-y", *apt_confdef, *apt_confold])
    else:
        logger.info("--skip-full-update option passed - update all currently installed packages skipped")

    logger.info("Installing required packages...")

    os_version = get_os_info()["VERSION_ID"]

    # Validate the version of the OS to match the format xx.yy
    if not re.match(r"^\d{2}\.\d{2}$", os_version):
        logger.error("Invalid VERSION_ID format in /etc/os-release: %s.", os_version)
        logger.error("Expected format: xx.yy (e.g. 20.04)")
        sys.exit(1)

    os_version = os_version.replace(".", "_")

    install_cl_release(os_version)
    if os.path.exists('/etc/cpanel/ea4/is_ea4'):
        install_ea4_release(os_version)

    # setup CL Build System repos, if any (for testing)
    setup_build_repos()

    run_command(["apt-get", "update"])

    # Install RHN packages and register the server
    run_command(["apt-get", "install", "-y", *RHN_PACKAGES])
    register_server()

    # Install all other required packages
    logger.info("Proceeding with the installation...")
    packages_to_install = get_packages_to_install()
    run_command(["apt-get", "install", "-y", *packages_to_install])

    logger.info("Performing reconfiguration of CloudLinux edition-specific features...")
    try:
        run_command(["/usr/sbin/rhn_check", "--allow-transition"])
    except (subprocess.CalledProcessError, SystemExit):
        log_message = "Failed to reconfigure CloudLinux edition-specific features"
        if not parser_args.skip_registration:
            logger.exception(log_message)
            sys.exit(1)
        else:
            # If the skip-registration option was passed, the failure is not critical,
            # because rhn_check depends on files that are created during registration.
            logger.warning(log_message)
            logger.warning("skip-registration option was passed, continuing with the conversion process")


def download_file(url: str, new_version: str) -> None:
    """
    Download file using tmp.

    Arguments:
        url: link to download file from
        new_version: used for logging

    """
    logger.info("Downloading file: %s", url)

    with tempfile.NamedTemporaryFile(dir="/root") as tmp:
        try:
            urlretrieve(url, tmp.name)
            # Change content of the ubuntu2cloudlinux script in a runtime
            shutil.copy(tmp.name, script_file)
        except URLError as err:
            logger.info("Error encountered while downloading and copying file %s %s: %s", script_name, new_version, err)
            sys.exit(1)

    logger.info("Download of %s %s successful", script_name, new_version)


def is_upgrade(new_version: str) -> bool:
    """
    Check if the user wants to download and run the new version of the script.

    Arguments:
        new_version: used for logging

    """
    while True:
        user_input = input(f"download and run {script_name} {new_version} now? [yes/no] ").lower()
        if user_input not in ("yes", "no", "y", "n"):
            continue
        break

    return user_input in ("yes", "y")


def spawn_child_process() -> None:
    """
    Run a new version of ubuntu2cloudlinux.py in a subprocess.

    Attaches stdout and stderr of the child process to parent stdout and stderr.
    """
    # getLogger() returns the root logger that was previously configured in setup_logger()
    # Handlers and formatters are preserved.
    logger = logging.getLogger()
    # Remove the logfile handler from the logger, as the child process will also write to it
    logger.handlers.pop()

    cmd = ["python3", script_file, "--skip-version-check", *sys.argv[1:]]

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # noqa: S603

    if p.stdout is None or p.stderr is None:
        logger.error("Failed to open stdout or stderr for the child process")
        sys.exit(1)

    # The child process will write to stdout and stderr, and we will read and display them here
    pipes = select.select([p.stdout.fileno(), p.stderr.fileno()], [], [])

    while p.poll() is None:
        for pipe in pipes[0]:
            if pipe == p.stdout.fileno():
                line = p.stdout.readline().decode().strip("\n")
                if len(line):
                    logger.info(line)

            elif pipe == p.stderr.fileno():
                line = p.stderr.readline().decode().strip("\n")
                if len(line):
                    logger.info(line)
                    logger.debug(line)

    sys.exit(p.returncode)


def is_latest_version(remote_version: str | None = None) -> bool:
    """Compare current version with input or from the repo."""
    logger.info("Checking for an update to %s", script_name)
    remote_version = remote_version or get_latest_version()
    if ScriptVersion(script_version) >= ScriptVersion(remote_version):
        logger.info("%s is already the latest version %s - continuing", script_name, script_version)
        return True
    return False


def get_latest_version() -> str:
    """Get the latest ubuntu2cloudlinux version string from the repo."""
    remote_version = urlopen(VERSION_FILE_URL)
    return remote_version.readline().decode().strip()


def check_version() -> None:
    """
    Check if the new version of the script is available.

    1. If new version exists, download and execute in subprocess;
    2. If not, continue with the current script.
    """
    remote_version = get_latest_version()
    if is_latest_version(remote_version):
        return

    logger.info("New version %s of %s detected", remote_version, script_name)
    if os.getenv("CLDEPLOY_TESTING"):
        pass  # do not wait for user input in test env
    elif not is_upgrade(remote_version):
        return

    download_file(SCRIPT_URL, remote_version)

    logger.info("Running new version of %s %s", script_name, remote_version)
    spawn_child_process()


def handle_general_error(
    exception_instance: Exception | SystemExit,
    console_output: str = "Errors occurred during conversion process.",
) -> None:
    """Log an error that stopped the conversion process and output the error message to the console."""
    if not isinstance(exception_instance, SystemExit):
        logger.critical("Uncaught error occurred during ubuntu2cloudlinux execution: %s", exception_instance)
    log_with_color(
        console_output,
        Color.RED,
        level=logging.ERROR,
    )


def generate_sysinfo_cpanel() -> int:
    """Update some cPanel data, like in cldeploy."""
    rc = 0
    if Path("/scripts/gensysinfo").exists():
        logger.info("Launching /scripts/gensysinfo in order to update cPanel sysinfo...")
        if Path("/var/cpanel/sysinfo.config").exists():
            Path("/var/cpanel/sysinfo.config").unlink()
        rc = run_command(["/scripts/gensysinfo"], exit_on_error=False)
    return rc


def uninstall_cloudlinux() -> None:
    """
    Perform uninstallation of CloudLinux Subsystem.

    Run uninstall scripts for existing CL components, remove main CL packages
    """
    failed_steps = []
    apt_remove_cmd = ["apt-get", "remove", "--purge", "-y"]
    if Path("/usr/sbin/cagefsctl").exists():
        logger.info("Erasing cagefs...")
        rc = run_command(["/usr/sbin/cagefsctl", "--do-not-ask", "--remove-all"], exit_on_error=False)
        rc += run_command([*apt_remove_cmd, "cagefs", "cagefs-safebin", "bsock", "bsock-libs"], exit_on_error=False)
        if rc > 0:
            failed_steps.append("Erasing cagefs")
    if Path("/usr/sbin/db_governor").exists() and Path("/usr/share/lve/dbgovernor/mysqlgovernor.py").exists():
        logger.info("Erasing db-governor...")
        rc = run_command(["/usr/share/lve/dbgovernor/mysqlgovernor.py", "--delete"], exit_on_error=False)
        if rc:
            failed_steps.append("Erasing db-governor")
    if glob("/opt/alt/php*/usr/bin/php"):
        logger.info("Erasing alt-php...")
        rc = run_command([*apt_remove_cmd, "alt-php*"], exit_on_error=False)
        if rc:
            failed_steps.append("Erasing alt-php")
    if glob("/opt/alt/python*/bin/python*"):
        logger.info("Erasing alt-python...")
        rc = run_command([*apt_remove_cmd, "alt-python*"], exit_on_error=False)
        if rc:
            failed_steps.append("Erasing alt-python")
    # Don't try to remove CL packages if cloudlinux-release was already removed
    # (cloudlinux.list is a part of the cloudlinux-release package)
    if Path("/etc/apt/sources.list.d/cloudlinux.list").exists():
        logger.info("Erasing main CloudLinux packages...")
        packages_to_uninstall = RHN_PACKAGES + COMMON_PACKAGES

        rc = run_command([*apt_remove_cmd, *packages_to_uninstall], exit_on_error=False)
        if rc:
            failed_steps.append("Erasing main CloudLinux packages")

    # remove meta packages
    package_marker_map = {
        CL_SOLO_META_PACKAGE: "/etc/cloudlinux-edition-solo",
        CL_ADMIN_META_PACKAGE: "/etc/cloudlinux-edition-admin",
        CL_CONTAINER_META_PACKAGE: "/etc/cloudlinux-container",
    }
    for package, marker in package_marker_map.items():
        if Path(marker).exists():
            logger.info("Erasing %s", package)
            rc = run_command([*apt_remove_cmd, package], exit_on_error=False)
            if rc:
                failed_steps.append(f"Erasing {package}")

    # remove cloudlinux-release only if all previous steps finished without errors
    # because user may need CL repo to fix some problems with packages
    if Path("/etc/apt/sources.list.d/cloudlinux.list").exists() and not failed_steps:
        logger.info("Erasing cloudlinux-release package")
        rc = run_command([*apt_remove_cmd, "cloudlinux-release"], exit_on_error=False)
        if rc:
            failed_steps.append("Erasing cloudlinux-release package")

    rc = generate_sysinfo_cpanel()
    if rc:
        failed_steps.append("Launching cpanel script /scripts/gensysinfo")

    if failed_steps:
        raise RuntimeError("Following actions finished with error: " + ", ".join(failed_steps))


def main() -> int:
    """
    Execute the main logic of the script.

    Perform Ubuntu conversion or uninstallation of the CL Subsystem.
    """
    if not parser_args.skipversioncheck:
        check_version()

    if parser_args.mode_uninstall:
        logger.info("Starting uninstallation of the CloudLinux Extension...")
        try:
            uninstall_cloudlinux()
        except (Exception, SystemExit) as e:
            error_output = (
                "Errors occurred during CloudLinux Subsystem uninstallation.\n"
                f"Please check the log, fix errors and restart '{script_name} --uninstall' if necessary"
            )
            handle_general_error(e, error_output)
            exit_code = 1
        else:
            log_with_color(
                [
                    "Uninstallation completed successfully!\n",
                    "Some packages from CloudLinux could remain installed because",
                    "their automatic deletion may break some functionality in your system.",
                    "You can find packages like that using the command",
                    "dpkg-query --show --showformat="
                    "'${Package}\\t${Homepage}\\t${Maintainer}\\t${Version}\\n' | grep -i cloudlinux",
                    "and manually replace them with upstream versions if ones exist in the upstream repositories.",
                    "You can also run 'apt autoremove' to get rid of all packages",
                    "that were installed as dependencies and no longer required.\n",
                    "Also you need to reboot the system to unload CloudLinux kernel modules.",
                ],
                Color.GREEN,
            )
            exit_code = 0

    else:
        logger.info("Starting Ubuntu conversion...")
        try:
            perform_ubuntu_conversion()
        except (Exception, SystemExit) as e:
            handle_general_error(e)
            exit_code = 1
        else:
            log_with_color(
                [
                    "Conversion completed successfully!",
                    "Please reboot the system to enable installed kernel modules.",
                ],
                Color.GREEN,
            )
            exit_code = 0

    logger.info("You can find the complete log in %s", LOG_PATH)
    return exit_code


if __name__ == "__main__":
    check_ubuntu()
    parser_args = setup_parser()
    sys.exit(main())
